"""
Acacia - a programming language that runs in Minecraft Bedrock Edition
by compiling code into mcfunction commands, supporting variables,
functions, custom types, arithmetic, entity and coordinate system and
very rich compile-time structures like strings, lists and maps.
"""

__version__ = "0.1.0"
